import math
import gurobipy
import ptn_data
import line_data
import od_data
from typing import List
from typing import Dict
from collections import defaultdict


class Ean:
    """
    Class representing the Event Activity Network
    """

    def __init__(self, ptn: ptn_data.Ptn = None, line_pool: line_data.LinePool = None, od: od_data.OD = None,
                 multiple_frequencies: bool = False, period_length: int = 60) -> None:
        """
        Initialisation of the event activity network from a given PTN.
        OD nodes and arcs as well as synchronization arcs are added as well
        :param ptn: corresponding PTN
        :param line_pool: corresponding line pool
        :param od: corresponding OD data
        :param multiple_frequencies: whether multiple frequencies are used
        :param period_length: corresponding period length
        """
        self.events = []
        self.first_events_in_line = {}
        self.last_events_in_line = {}
        if ptn is not None and line_pool is not None and od is not None:
            self.init_ptn(ptn, line_pool)
            self.init_od(ptn, od)
            if multiple_frequencies:
                self.init_sync(period_length)

    def add_event(self, event: "EanEvent") -> None:
        if not self.events.__contains__(event):
            self.events.append(event)

    def get_all_events(self) -> List["EanEvent"]:
        return self.events

    def get_events_network(self) -> List["EanEventNetwork"]:
        events = []
        for event in self.events:
            if isinstance(event, EanEventNetwork):
                events.append(event)
        return events

    def get_activities(self, activity_type_list: list) -> List["EanActivity"]:
        activities = []
        for event in self.events:
            for activity in event.get_activities():
                if activity.get_activity_type() in activity_type_list and activity not in activities:
                    activities.append(activity)
        return activities

    def get_all_activities(self) -> List["EanActivity"]:
        activities = []
        for event in self.events:
            for activity in event.get_activities():
                if activity not in activities:
                    activities.append(activity)
        return activities

    def get_event_od(self, start: ptn_data.PtnNode, end: ptn_data.PtnNode, od_event_type: str, time_1: int,
                     time_2: int) -> "EanEventOD":
        for event in self.events:
            if isinstance(event, EanEventOD) and event.check_attributes(start, end, od_event_type, time_1, time_2):
                return event

    def get_dict_events_to_ptn_nodes(self, ptn: ptn_data.Ptn, event_type: str) -> Dict[
                                                                                    "PtnNode", List["EanEventNetwork"]]:
        events_to_ptn_node = {}
        for ptn_node in ptn.get_nodes():
            events_to_ptn_node[ptn_node] = []
        for event in self.events:
            if isinstance(event, EanEventNetwork) and event.get_event_type() == event_type:
                events_to_ptn_node[event.get_ptn_node()].append(event)
        return events_to_ptn_node

    def get_event_by_event_id(self, event_id: int) -> "EanEvent":
        for event in self.events:
            if event.get_event_id() == event_id:
                return event
        return None

    def init_ptn(self, ptn: ptn_data.Ptn, line_pool: line_data.LinePool) -> None:
        """
        Initialisation of the part of the EAN which correspond to the PTN and the line pool.
        Arrival and departure events are initialized as well as drive, wait and transfer activities.
        :param ptn: corresponding PTN
        :param line_pool: corresponding line pool
        :return: None
        """
        for line in line_pool.get_lines():
            first = True
            for edge in line.get_edges():
                left_node = edge.get_left_node()
                right_node = edge.get_right_node()
                if first:
                    # First Wait-Activity
                    departure = EanEventNetwork(self, left_node, line, 'dep')
                    self.first_events_in_line[line] = departure
                    first = False
                # Drive-Activity
                arrival = EanEventNetwork(self, right_node, line, 'arr')
                EanActivity(departure, arrival, 'drive', edge.get_min_drive_time(), edge.get_max_drive_time())
                self.last_events_in_line[line] = arrival
                # Wait-Activity
                if right_node != line.get_last_stop():
                    departure = EanEventNetwork(self, right_node, line, 'dep')
                    EanActivity(arrival, departure, 'wait', left_node.get_min_wait_time(),
                                left_node.get_max_wait_time())
        # Change-Activities
        arr_events_to_ptn_node = self.get_dict_events_to_ptn_nodes(ptn, 'arr')
        dep_events_to_ptn_node = self.get_dict_events_to_ptn_nodes(ptn, 'dep')
        for ptn_node in ptn.get_nodes():
            lower_bound = ptn_node.get_min_trans_time()
            upper_bound = ptn_node.get_max_trans_time()
            for arr_event in arr_events_to_ptn_node[ptn_node]:
                for dep_event in dep_events_to_ptn_node[ptn_node]:
                    if arr_event.get_directed_line_id() != dep_event.get_directed_line_id():
                        EanActivity(arr_event, dep_event, 'trans', lower_bound, upper_bound)

    def init_od(self, ptn: ptn_data.Ptn, od: od_data.OD) -> None:
        """
        Initialisation of the OD part of the EAN.
        Source and target events for all OD pairs and time slices are constructed as well as arcs to and from the
        network as well as time arcs which are used for changing the time slice in which a journey begins.
        :param ptn: corresponding PTN
        :param od: corresponding OD data
        :return: None
        """
        arr_events_to_ptn_node = self.get_dict_events_to_ptn_nodes(ptn, 'arr')
        dep_events_to_ptn_node = self.get_dict_events_to_ptn_nodes(ptn, 'dep')
        for od_pair in od.get_all_od_pairs():
            if od_pair.get_n_passengers(1) == 0:
                continue
            origin = od_pair.get_origin()
            destination = od_pair.get_destination()
            n_time_slices = od_pair.get_n_time_slices()
            dep_events_origin = dep_events_to_ptn_node[origin]
            arr_events_destination = arr_events_to_ptn_node[destination]
            for t in range(1, n_time_slices + 1):
                # A time
                start_event = EanEventOD(self, origin, destination, 'source', t, t)
                for event in dep_events_origin:
                    EanActivity(start_event, event, 'to', None, None)
                for t_2 in range(1, n_time_slices + 1):
                    if t != t_2:
                        event_t_2 = EanEventOD(self, origin, destination, 'source', t, t_2)
                        EanActivity(start_event, event_t_2, 'time', None, None)
                        # A to
                        for event in dep_events_origin:
                            EanActivity(event_t_2, event, 'to', None, None)
                # A from
                end_event = EanEventOD(self, origin, destination, 'target', t)
                for event in arr_events_destination:
                    EanActivity(event, end_event, 'from', None, None)

    def init_sync(self, period_length: int) -> None:
        """
        Initialisation of synchronization activities between events belonging to different frequency repetitions of
        the same line.
        :param period_length: corresponding period length
        :return: None
        """
        network_events = self.get_events_network()
        for event_1 in network_events:
            line_1 = event_1.get_line()
            line_id_1 = line_1.get_directed_line_id()
            frequency_1 = line_1.get_frequency()
            repetition_1 = line_1.get_repetition()
            if frequency_1 > 1 and repetition_1 == 1:
                events_to_sync = {}
                events_to_sync[1] = event_1
                for event_2 in network_events:
                    line_2 = event_2.get_line()
                    line_id_2 = line_2.get_directed_line_id()
                    repetition_2 = line_2.get_repetition()
                    if line_id_1 == line_id_2 and repetition_2 != 1 \
                            and event_1.get_ptn_node() == event_2.get_ptn_node() \
                            and event_1.get_event_type() == event_2.get_event_type():
                        events_to_sync[repetition_2] = event_2
                for repetition in range(1, frequency_1):
                    EanActivity(events_to_sync[repetition], events_to_sync[repetition + 1], 'sync',
                                int(period_length / frequency_1), int(period_length / frequency_1))

    def compute_max_upper_bound(self) -> int:
        """
        Computation of the maximal upper bound of a wait, drive or transfer activity.
        :return: maximal upper bound of a drive, wait or transfer activity.
        """
        max_upper_bound = 0
        for activity in self.get_activities(['wait', 'drive', 'trans']):
            if activity.get_upper_bound() > max_upper_bound:
                max_upper_bound = activity.get_upper_bound()
        return max_upper_bound


class EanActivity:
    """
    Class representing an activity, i.e., an arc of the EAN
    """

    def __init__(self, left_event: "EanEvent", right_event: "EanEvent", activity_type: str, lower_bound: int,
                 upper_bound: int, n_passengers: int = 0, activity_id: int = -1) -> None:
        """
        Initialization of an activity, i.e., an arc of the EAN.
        :param left_event: source event
        :param right_event: target event
        :param activity_type: type of the activity ("drive", "wait", "trans", "sync", "to", "from", "time")
        :param lower_bound: lower bound on the duration of the activity
        :param upper_bound: upper bound on the duration of the activity
        :param n_passengers: number of passengers using the activity
        :param activity_id: id of the activity
        """
        self.left_event = left_event
        self.right_event = right_event
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.activity_type = activity_type
        left_event.add_activity(self)
        self.activity_id = activity_id
        self.n_passengers = n_passengers

    def set_activity_id(self, activity_id: int) -> None:
        self.activity_id = activity_id

    def get_activity_id(self) -> int:
        return self.activity_id

    def get_activity_type(self) -> str:
        return self.activity_type

    def get_lower_bound(self) -> int:
        return self.lower_bound

    def get_upper_bound(self) -> int:
        return self.upper_bound

    def get_n_passengers(self) -> int:
        return self.n_passengers

    def set_n_passengers(self, n_passengers: int) -> None:
        self.n_passengers = n_passengers

    def add_passengers(self, n_passengers: int) -> None:
        self.n_passengers += n_passengers

    def to_string(self) -> str:
        return "(" + self.left_event.to_string() + "," + self.right_event.to_string() + "," + self.activity_type + ")"

    def get_left_event(self) -> "EanEvent":
        return self.left_event

    def get_right_event(self) -> "EanEvent":
        return self.right_event

    def to_activities_periodic(self) -> str:
        if self.activity_type == 'drive':
            type_string = '\"drive\"'
        elif self.activity_type == 'wait':
            type_string = '\"wait\"'
        elif self.activity_type == 'sync':
            type_string = '\"sync\"'
        elif self.activity_type == 'trans':
            type_string = '\"change\"'
        else:
            type_string = '\"%s\"' % self.activity_type
        return "%d; %s; %d; %d; %d; %d;" % (
            self.activity_id, type_string, self.left_event.get_event_id(), self.right_event.get_event_id(),
            self.lower_bound,
            self.upper_bound)


class EanEvent:
    """
    Class representing an event, i.e., a node of the EAN
    """

    def __init__(self, ean: Ean, event_type: str) -> None:
        """
        Initialisation of an event, i.e., a node in the EAN
        :param ean: corresponding EAN
        :param event_type: type of the event ("dep": departure, "arr": arrival, "source", "target")
        """
        self.incident_activities = []
        self.event_type = event_type
        ean.add_event(self)
        self.event_id = -1

    def add_activity(self, activity: EanActivity) -> None:
        if not self.incident_activities.__contains__(activity):
            self.incident_activities.append(activity)

    def get_activities(self) -> List[EanActivity]:
        return self.incident_activities

    def get_event_type(self) -> str:
        return self.event_type

    def get_event_id(self) -> int:
        return self.event_id

    def set_event_id(self, event_id: int) -> None:
        self.event_id = event_id


class EanEventNetwork(EanEvent):
    """
    class representing a network event, i.e., an event that is not linked to a specified OD pair
    """

    def __init__(self, ean: Ean, ptn_node: ptn_data.PtnNode, line: line_data.Line, event_type: str,
                 n_passengers: int = 0, event_id: int = -1) -> None:
        """
        Initialisation of a netwrok event
        :param ean: corresponding EAN
        :param ptn_node: corresponding PTN node
        :param line: corresponding line
        :param event_type: type of the event ("arr": arrival, "dep": departure)
        :param n_passengers: number of passenger using the event
        :param event_id: id of the event
        """
        self.incident_activities = []
        ean.add_event(self)
        self.ptn_node = ptn_node
        self.line = line
        self.event_type = event_type  # arr or dep
        self.event_id = event_id
        self.n_passengers = n_passengers
        self.event_time = -1

    def get_n_passengers(self) -> int:
        return self.n_passengers

    def set_n_passengers(self, n_passengers: int) -> None:
        self.n_passengers = n_passengers

    def add_passengers(self, n_passengers: int) -> None:
        self.n_passengers += n_passengers

    def set_event_id(self, event_id: int) -> None:
        self.event_id = event_id

    def get_event_id(self) -> int:
        return self.event_id

    def get_ptn_node(self) -> ptn_data.PtnNode:
        return self.ptn_node

    def get_directed_line_id(self) -> int:
        return self.line.get_directed_line_id()

    def get_line(self) -> line_data.Line:
        return self.line

    def get_event_type(self) -> str:
        return self.event_type

    def check_attributes(self, ptn_node: ptn_data.PtnNode, line: line_data.Line, event_type: str) -> bool:
        return self.ptn_node.get_node_id() == ptn_node.get_node_id() and self.line == line and self.event_type == event_type

    def to_string(self) -> str:
        return "(" + self.ptn_node.to_string() + "," + self.line.to_string() + "," + self.event_type + ")"

    def to_events_periodic(self) -> str:
        if self.event_type == 'arr':
            type_string = '\"arrival\"'
        else:
            type_string = '\"departure\"'
        return "%d; %s; %d; %d; %d; %s; %d" % (
            self.event_id, type_string, self.ptn_node.get_node_id(), math.floor(self.line.get_undirected_line_id()),
            self.get_n_passengers(), self.get_line().get_direction(), self.get_line().get_repetition())


class EanEventOD(EanEvent):
    """
    Class representing the OD events corresponding to specified OD pairs which are used as source and target for their
    journeys.
    """
    def __init__(self, ean: Ean, start: ptn_data.PtnNode, end: ptn_data.PtnNode, od_event_type: str, time_1: int,
                 time_2: int = None, event_id: int = -1) -> None:
        """
        Initialization of an OD event.
        :param ean: corresponding EAN
        :param start: PTN source node of the corresponding OD pair
        :param end: PTN target node of the corresponding OD pair
        :param od_event_type: type of the event ("source", "target")
        :param time_1: time slice in which the journey of the OD pair is supposed to start
        :param time_2: time slice to which the event belongs or None if the event is a target event
        :param event_id: id of the event
        """
        self.incident_activities = []
        ean.add_event(self)
        self.start = start
        self.end = end
        self.od_event_type = od_event_type
        self.time_1 = time_1
        self.time_2 = time_2  # none, if type=target
        self.event_id = event_id

    def get_start(self) -> ptn_data.PtnNode:
        return self.start

    def get_end(self) -> ptn_data.PtnNode:
        return self.end

    def get_time_1(self) -> int:
        return self.time_1

    def get_time_2(self) -> int:
        return self.time_2

    def check_attributes(self, start: ptn_data.PtnNode, end: ptn_data.PtnNode, od_event_type: str, time_1: int,
                         time_2: int) -> bool:
        return self.start == start and self.end == end and self.od_event_type == od_event_type and self.time_1 == time_1 and self.time_2 == time_2

    def check_attributes_od(self, start: ptn_data.PtnNode, end: ptn_data.PtnNode, time_1: int) -> bool:
        return self.start == start and self.end == end and self.time_1 == time_1

    def to_string(self) -> str:
        if self.time_2 is None:
            return "(" + self.start.to_string() + "," + self.end.to_string() + "," + str(
                self.time_1) + "," + self.od_event_type + ")"
        else:
            return "(" + self.start.to_string() + "," + self.end.to_string() + "," + str(
                self.time_1) + "," + str(self.time_2) + "," + self.od_event_type + ")"
