import gurobipy
import ean_data
import ptn_data
import line_data
import od_data
import preprocessing_tim_pass
from cycle_base import *
import math


def solve_cycle_base(ean: ean_data.Ean, ptn: ptn_data.Ptn, line_pool: line_data.LinePool, od: od_data.OD,
                     period_length: int, activity_file_name: str, event_file_name: str, timetable_file_name: str,
                     factor_travel_time: float, factor_drive_time: float, factor_transfer_time: float,
                     factor_wait_time: float, factor_penalty_transfer: float, factor_penalty_time_slice: float,
                     tim_pass_fix_passengers: bool, use_preprocessing: bool, time_limit: int, mip_gap: float,
                     write_lp_output: bool = False) -> None:
    '''
    Method to construct and solve the cycle based IP model.
    :param ean: the event activity network
    :param ptn: the public transprotation network
    :param line_pool: the line pool
    :param od: the OD matrix
    :param period_length:  the period length
    :param time_limit: the time limit for the optimization
    :param tim_pass_fix_passengers: whether Heuristic UB (True) or LB (False) should be used
    :param mip_gap: the maximal gap for the optimization problem
    :param activity_file_name: file name activites
    :param event_file_name: file name events
    :param timetable_file_name: file name timetable
    :param factor_travel_time: weight for travel time
    :param factor_drive_time: weight for drive time
    :param factor_transfer_time: weight for transfer time
    :param factor_wait_time: weight for wait time
    :param use_preprocessing: whether preprocessing should be used
    :param factor_penalty_time_slice: penalty weigth for changing time slices
    :param factor_penalty_transfer: penalty weigth for transfers
    :param write_lp_output: whether the IP model is written to a file
    :return: None
    '''

    unused_transfer_activities = []
    # Preprocessing
    print("Compute fixed passenger paths...")
    preprocessing_tp = preprocessing_tim_pass.PreprocessingTimPass(ean, factor_travel_time, factor_drive_time,
                                                               factor_transfer_time,
                                                               factor_wait_time, factor_penalty_transfer,
                                                               unused_transfer_activities)
    if tim_pass_fix_passengers:
         preprocessing_tp.compute_fixed_passenger_paths(ean, od, True)
    print("done!")

    # Big M constraints
    max_upper_bound = ean.compute_max_upper_bound()

    m_2 = period_length
    m_4 = max_upper_bound

    print("m_2=%d\nm_4=%d\n" % (m_2, m_4))

    # Lists for loops
    activities_drive_wait_trans = ean.get_activities(['drive', 'wait', 'trans'])
    activities_drive_wait_trans_sync = ean.get_activities(['drive', 'wait', 'trans', 'sync'])
    activities_time_to = ean.get_activities(['time', 'to'])
    activities_from = ean.get_activities(['from'])
    activities_no_sync = ean.get_activities(['drive', 'wait', 'trans', 'time', 'to', 'from'])
    active_od_pairs = od.get_active_od_pairs()

    print("Number of drive, wait, transfer activities: %d" % activities_drive_wait_trans.__len__())
    print("Number of drive, wait, transfer activities without unused transfer activities: %d" % (
        activities_drive_wait_trans.__len__() - unused_transfer_activities.__len__()))

    # Dictionaries for storing variables
    tension = {}
    modulo_parameter = {}
    arc_used = {}
    travel_time_linear = {}

    # Initialize Model
    print("Initialize model")
    m = gurobipy.Model("TimPass")
    if time_limit != -1:
        m.params.timeLimit = time_limit
    if mip_gap >= 0:
        m.params.MIPGap = mip_gap
    m.modelSense = gurobipy.GRB.MINIMIZE
    print("Calculating cycle-base")
    cb = cycle_base(ean)
    # Initialize Variables
    print("Initialize variables")

    # tensions
    print("\ttension")
    for activity in activities_drive_wait_trans_sync:
        tension[activity] = m.addVar(vtype=gurobipy.GRB.INTEGER, name='tension_%s' % activity.to_string(),
                                     lb=activity.lower_bound, ub=activity.upper_bound)
    # modulo parameter
    print("\tmodulo parameter")
    for non_tree_activity in cb:
        LB = 0; UB = 0
        for activity, orientation in cb[non_tree_activity]:
            if orientation == 1:
                UB += activity.upper_bound
                LB += activity.lower_bound
            else:
                UB -= activity.lower_bound
                LB -= activity.upper_bound
        UB = math.floor(UB/(period_length*1.0))
        LB = math.ceil(LB/(period_length*1.0))
        modulo_parameter[non_tree_activity] = m.addVar(lb=LB, ub=UB, vtype=gurobipy.GRB.INTEGER,
                                                       name='z_%s' % non_tree_activity.to_string())

    # Arc used
    print("\tarc used")
    for activity in activities_no_sync:
        arc_used[activity] = {}
        for od_pair in active_od_pairs:
            arc_used[activity][od_pair] = {}
            for t in range(1, od_pair.get_n_time_slices() + 1):
                arc_used[activity][od_pair][t] = m.addVar(vtype=gurobipy.GRB.BINARY, name='p_%s_%s_%d' % (
                    activity.to_string(), od_pair.to_string(), t))

    # Travel time linear
    print("\ttravel time linear")
    for activity in activities_drive_wait_trans:
        travel_time_linear[activity] = {}
        for od_pair in active_od_pairs:
            travel_time_linear[activity][od_pair] = {}
            for t in range(1, od_pair.get_n_time_slices() + 1):
                travel_time_linear[activity][od_pair][t] = m.addVar(vtype=gurobipy.GRB.INTEGER, name='d_%s_%s_%d' % (
                    activity.to_string(), od_pair.to_string(), t))

    # Add variables to model m
    m.update()
    print("Number of variables: " + str(m.getAttr(gurobipy.GRB.Attr.NumVars)))

    # objective function

    print("Initialize objective function")
    sum_travel_time = 0
    sum_drive_time = 0
    sum_wait_time = 0
    sum_transfer_time = 0
    sum_penalty_changing_time_slices = 0
    sum_transfers = 0

    for od_pair in active_od_pairs:
        for t in range(1, od_pair.get_n_time_slices() + 1):
            for activity in activities_drive_wait_trans:
                sum_travel_time += travel_time_linear[activity][od_pair][t] * od_pair.get_n_passengers(t)
                if activity.get_activity_type() == 'trans':
                    sum_transfers += arc_used[activity][od_pair][t] * od_pair.get_n_passengers(t)
                    sum_transfer_time += travel_time_linear[activity][od_pair][t] * od_pair.get_n_passengers(t)
                if activity.get_activity_type() == 'wait':
                    sum_wait_time += travel_time_linear[activity][od_pair][t] * od_pair.get_n_passengers(t)
                if activity.get_activity_type() == 'drive':
                    sum_drive_time += travel_time_linear[activity][od_pair][t] * od_pair.get_n_passengers(t)
            for activity in ean.get_activities(['time']):
                time_2 = activity.get_right_event().get_time_2()
                sum_penalty_changing_time_slices += od_pair.get_penalty(t, time_2) * period_length * \
                                                    arc_used[activity][od_pair][t] * od_pair.get_n_passengers(t)

    # add fix passenger paths
    for activity in activities_drive_wait_trans:
        if activity.get_n_passengers() == 0:
            continue
        sum_travel_time += tension[activity] * activity.get_n_passengers()
        if activity.get_activity_type() == 'trans':
            sum_transfers += activity.get_n_passengers()
            sum_transfer_time += tension[activity] * activity.get_n_passengers()
        if activity.get_activity_type() == 'wait':
            sum_wait_time += tension[activity] * activity.get_n_passengers()
        if activity.get_activity_type() == 'drive':
            sum_drive_time += tension[activity] * activity.get_n_passengers()

    m.setObjective(factor_travel_time * sum_travel_time +
                   factor_drive_time * sum_drive_time + factor_transfer_time * sum_transfer_time +
                   factor_wait_time * sum_wait_time +
                   factor_penalty_time_slice * sum_penalty_changing_time_slices +
                   factor_penalty_transfer * sum_transfers)

    # Add constraints
    print("Add constraints:")

    # cycle constraints
    cycle_constraint = []
    for non_tree_activity in cb:
        con = gurobipy.LinExpr()
        for activity, orientation in cb[non_tree_activity]:
            con += tension[activity] * orientation
        constraint = m.addConstr(con == period_length * modulo_parameter[non_tree_activity], name='cycle_constraint_%s' % non_tree_activity.to_string())
        cycle_constraint.append(constraint)

    # Passenger flow
    print("\tpassenger flow")
    passenger_flow_constraints = []
    passenger_flow_constraints_by_od_pair = {}
    counter = 1
    for od_pair in active_od_pairs:
        print("\t\tOd-pair: %d" % counter)
        counter += 1

        passenger_flow_constraints_by_od_pair[od_pair] = []

        if use_preprocessing:
            used_tuple = preprocessing_tp.compute_potentially_used_events_and_activities(ean, od_pair)
            used_events = used_tuple[0]
            used_activities_drive_wait_trans = used_tuple[1]
        else:
            used_events = ean.get_all_events()
            used_activities_drive_wait_trans = activities_drive_wait_trans

        used_activities_drive_wait_trans = set(used_activities_drive_wait_trans) - set(unused_transfer_activities)

        for t in range(1, od_pair.get_n_time_slices() + 1):
            for event in used_events:
                sum_out = 0
                sum_in = 0
                right_hand_side = 0
                empty = True
                activities_to_check = []
                for activity in used_activities_drive_wait_trans:
                    activities_to_check.append(activity)
                for activity in activities_time_to:
                    event_to_check_od = activity.get_left_event()
                    event_to_check_used = activity.get_right_event()
                    if not use_preprocessing or event_to_check_used in used_events:
                        if event_to_check_od.check_attributes_od(od_pair.get_origin(), od_pair.get_destination(), t):
                            activities_to_check.append(activity)
                for activity in activities_from:
                    event_to_check_od = activity.get_right_event()
                    event_to_check_used = activity.get_left_event()
                    if not use_preprocessing or event_to_check_used in used_events:
                        if event_to_check_od.check_attributes_od(od_pair.get_origin(), od_pair.get_destination(), t):
                            activities_to_check.append(activity)
                for activity in activities_to_check:
                    if activity.get_left_event() == event:
                        sum_out += arc_used[activity][od_pair][t]
                        empty = False
                    if activity.get_right_event() == event:
                        sum_in += arc_used[activity][od_pair][t]
                        empty = False
                if isinstance(event, ean_data.EanEventOD):
                    if event.check_attributes(od_pair.get_origin(), od_pair.get_destination(), 'source', t, t):
                        right_hand_side = 1
                    elif event.check_attributes(od_pair.get_origin(), od_pair.get_destination(), 'target', t, None):
                        right_hand_side = -1
                if not empty:
                    const_1 = m.addConstr(sum_out - sum_in == right_hand_side,
                                          name="pass_routing_%s_%d_%s" % (od_pair.to_string(), t, event.to_string()))
                    if write_lp_output:
                        passenger_flow_constraints.append(const_1)
                        passenger_flow_constraints_by_od_pair[od_pair].append(const_1)

    # Linearization travel time
    print("\tlinearization travel time")
    linearization_travel_time_constraints = []
    for activity in activities_drive_wait_trans:
        for od_pair in od.get_active_od_pairs():
            for t in range(1, od_pair.get_n_time_slices() + 1):
                const_1 = m.addConstr(travel_time_linear[activity][od_pair][t] >= 0,
                                      name="linearization_1_%s_%d_%s" % (od_pair.to_string(), t, activity.to_string()))
                const_2 = m.addConstr(travel_time_linear[activity][od_pair][t] >=
                                      tension[activity] - (1 - arc_used[activity][od_pair][t]) * m_4,
                                      name="linearization_2_%s_%d_%s" % (od_pair.to_string(), t, activity.to_string()))
                if write_lp_output:
                    linearization_travel_time_constraints.append(const_1)
                    linearization_travel_time_constraints.append(const_2)

    # write lp-file
    if write_lp_output:
        m.update()
        m.write("TimPass.lp")

        dec_file = open('TimPass.dec', 'w')
        dec_file.write("PRESOLVED\n0\n")
        dec_file.write("NBLOCKS\n3\n")
        # Block 1: Timetabling
        dec_file.write("BLOCK 1\n")
        for const in cycle_constraint:
            dec_file.write(const.ConstrName + "\n")
        # Block 2: Passenger flow
        dec_file.write("BLOCK 2\n")
        for const in passenger_flow_constraints:
            dec_file.write(const.ConstrName + "\n")
        # Master-constraints: Coupling
        dec_file.write("MASTERCONSS\n")
        # Timetabling + routing
        for const in time_slice_constraints:
            dec_file.write(const.ConstrName + "\n")
        for const in linearization_travel_time_constraints:
            dec_file.write(const.ConstrName + "\n")
        dec_file.close()

        dec_file = open('TimPassByODPair.dec', 'w')
        n_blocks = 2 + passenger_flow_constraints_by_od_pair.__len__()
        dec_file.write("PRESOLVED\n0\n")
        dec_file.write("NBLOCKS\n%d\n" % n_blocks)
        # Block 1: Timetabling
        dec_file.write("BLOCK 1\n")
        for const in cycle_constraint:
            dec_file.write(const.ConstrName + "\n")
        # Block 2: Passenger flow
        block_number = 2
        for od_pair in active_od_pairs:
            dec_file.write("BLOCK %d\n" % block_number)
            block_number += 1
            for const in passenger_flow_constraints_by_od_pair[od_pair]:
                dec_file.write(const.ConstrName + "\n")
        # Master-constraints: Coupling
        dec_file.write("MASTERCONSS\n")
        # Timetabling + routing
        for const in time_slice_constraints:
            dec_file.write(const.ConstrName + "\n")
        for const in linearization_travel_time_constraints:
            dec_file.write(const.ConstrName + "\n")
        dec_file.close()

    # Optimization
    print("Start optimization")
    m.optimize()
    gap = 'gap; ' + str(m.MIPGap)
    solver_time = 'solver_time; ' + str(m.Runtime)
    a = open('statistic/solver_statistic.sta', 'w')
    a.write(solver_time + '\n')
    a.write(gap + '\n')
    a.close()
    print("End optimization")

    # Check if a feasible solution was found
    if m.SolCount == 0:
        print("No feasible solution was found!")
        return
    else:
        print("Feasible solution was found!")

    print("Print events")

    events_file = open(event_file_name, 'w')
    events_file.write("# event_index; type; stop; line; passengers; direction; repetition\n")
    event_index = 1
    for event in ean.get_events_network():
        passengers = event.get_n_passengers()
        for activity in event.get_activities():
            for od_pair in od.get_active_od_pairs():
                for t in range(1, od_pair.get_n_time_slices()):
                    if arc_used[activity][od_pair][t].x == 1:
                        passengers += od_pair.get_n_passengers(t)
        event.set_event_id(event_index)
        event.set_n_passengers(passengers)
        events_file.write("%s\n" % event.to_events_periodic())
        event_index += 1
    events_file.close()

    print("Print timetable")
    tension_solution = {}
    for activity in tension:
        tension_solution[activity] = int(tension[activity].x)
    pi = {}
    used = defaultdict(lambda: False)

    for event in ean.get_events_network():
        if used[event]:
            continue
        pi[event] = 0
        used[event] = True
        queue = [event]
        while len(queue) > 0:
            cur = queue.pop(0)
            for activity in cur.incident_activities:
                if activity in activities_drive_wait_trans:
                    if cur == activity.left_event:
                        if not used[activity.right_event]:
                            queue.append(activity.right_event)
                            used[activity.right_event] = True
                            pi[activity.right_event] = (pi[cur] + tension_solution[activity]) % period_length
                    elif cur == activity.right_event:
                        if not used[activity.left_event]:
                            queue.append(activity.left_event)
                            used[activity.left_event] = True
                            pi[activity.left_event] = (pi[cur] - tension_solution[activity]) % period_length

    timetable_file = open(timetable_file_name, 'w')
    timetable_file.write("# event_index; time\n")
    for event in ean.get_events_network():
        timetable_file.write("%d; %d\n" % (event.get_event_id(), pi[event]))
    timetable_file.close()

    print("Print activities")

    activities_file = open(activity_file_name, 'w')
    activities_file.write("# activity_index; type; from_event; to_event; lower_bound; upper_bound; passengers\n")
    activity_index = 1
    for activity in ean.get_activities(['drive', 'wait']):
        passengers = activity.get_n_passengers()
        for od_pair in od.get_active_od_pairs():
            for t in range(1, od_pair.get_n_time_slices() + 1):
                if arc_used[activity][od_pair][t].x == 1:
                    passengers += od_pair.get_n_passengers(t)
        activity.set_activity_id(activity_index)
        activity.set_n_passengers(passengers)
        activities_file.write("%s %d\n" % (activity.to_activities_periodic(), passengers))
        activity_index += 1
    for activity in ean.get_activities(['sync']):
        passengers = activity.get_n_passengers()
        activity.set_activity_id(activity_index)
        activities_file.write("%s %d\n" % (activity.to_activities_periodic(), passengers))
        activity_index += 1
    for activity in ean.get_activities(['trans']):
        passengers = activity.get_n_passengers()
        for od_pair in od.get_active_od_pairs():
            for t in range(1, od_pair.get_n_time_slices() + 1):
                if arc_used[activity][od_pair][t].x == 1:
                    passengers += od_pair.get_n_passengers(t)
        activity.set_activity_id(activity_index)
        activity.set_n_passengers(passengers)
        activities_file.write("%s %d\n" % (activity.to_activities_periodic(), passengers))
        activity_index += 1
    activities_file.close()