import line_data
import ptn_data
import od_data
from ean_data import *
import gurobipy
import networkx as nx


class PreprocessingTimPass:
    '''
    Class used for preprocessing.
    '''
    ean_graph = nx.DiGraph()
    reverse_graph = nx.DiGraph()
    activities_drive_wait_trans = []
    unused_trans_activities = []
    unused_events = None
    unused_activities = None
    length_sp_upper_bound = -1

    def __init__(self, ean: Ean, factor_travel_time: float, factor_drive_time: float, factor_transfer_time: float,
                 factor_wait_time: float, penalty_transfer: float,
                 unused_transfer_activities: List["EanActivity"] = []) -> None:

        self.activities_drive_wait_trans = ean.get_activities(['drive', 'wait', 'trans'])

        max_event_id = 0
        max_activity_id = 0
        PreprocessingTimPass.unused_trans_activities = unused_transfer_activities

        for event in ean.get_all_events():
            max_event_id += 1
            event.set_event_id(max_event_id)
            self.ean_graph.add_node(event.get_event_id())

        for activity in ean.get_all_activities():
            max_activity_id += 1
            activity.set_activity_id(max_activity_id)
            if activity in unused_transfer_activities:
                continue
            if activity.get_activity_type() in ['sync', 'headway', 'time']:
                continue
            elif activity.get_activity_type() == 'drive':
                lower_bound = (factor_travel_time + factor_drive_time) * activity.get_lower_bound()
                upper_bound = (factor_travel_time + factor_drive_time) * activity.get_upper_bound()
            elif activity.get_activity_type() == 'wait':
                lower_bound = (factor_travel_time + factor_wait_time) * activity.get_lower_bound()
                upper_bound = (factor_travel_time + factor_wait_time) * activity.get_upper_bound()
            elif activity.get_activity_type() == 'trans':
                lower_bound = (factor_travel_time + factor_transfer_time) * activity.get_lower_bound() \
                              + penalty_transfer
                upper_bound = (factor_travel_time + factor_transfer_time) * activity.get_upper_bound() \
                              + penalty_transfer
            elif activity.get_activity_type() in ['to', 'from']:
                lower_bound = 0
                upper_bound = 0
            self.ean_graph.add_edge(activity.get_left_event().get_event_id(), activity.get_right_event().get_event_id(),
                                    lower_bound=lower_bound, upper_bound=upper_bound)

        self.reverse_graph = self.ean_graph.reverse()

    def compute_fixed_passenger_paths(self, ean: Ean, od: od_data.OD, use_lower_bounds: bool = True) -> float:
        if use_lower_bounds:
            edge_weight = 'lower_bound'
        else:
            edge_weight = 'upper_bound'

        length_fix_passengers = 0

        for event in ean.get_events_network():
            event.set_n_passengers(0)

        for activity in self.activities_drive_wait_trans:
            activity.set_n_passengers(0)

        for od_pair in od.get_all_od_pairs():
            if od_pair.is_active() or od_pair.get_n_passengers(1) == 0:
                continue
            start_event = ean.get_event_od(od_pair.get_origin(), od_pair.get_destination(), 'source', 1, 1)
            end_event = ean.get_event_od(od_pair.get_origin(), od_pair.get_destination(), 'target', 1, None)
            #shortest_path = nx.dijkstra_path(self.ean_graph, start_event.get_event_id(), end_event.get_event_id(),
            #                                 edge_weight)
            lengths, shortest_paths = nx.single_source_dijkstra(self.ean_graph, start_event.get_event_id(),
                                                                target=end_event.get_event_id(), weight=edge_weight)
            shortest_path = shortest_paths[end_event.get_event_id()]
            length = lengths[end_event.get_event_id()]
            length_fix_passengers += length * od_pair.get_n_passengers(1)
            # add passenger weight to events and activities
            # first and last activities and events are excluded, as they belong to to and from arcs
            last_event_id = shortest_path[0]
            last_event = ean.get_event_by_event_id(last_event_id)
            for event_id in shortest_path[1: -1]:
                event = ean.get_event_by_event_id(event_id)
                if isinstance(event, EanEventNetwork):
                    event.add_passengers(od_pair.get_total_passengers())
                    # Don't check first activity
                    if not isinstance(last_event, EanEventNetwork):
                        last_event = event
                        continue
                    for outgoing_activity in last_event.get_activities():
                        if outgoing_activity.get_right_event().get_event_id() == event_id:
                            outgoing_activity.add_passengers(od_pair.get_total_passengers())
                last_event = event
        return length_fix_passengers

    def compute_potentially_used_events_and_activities(self, ean: Ean, od_pair: od_data.ODPair) -> list:
        self.unused_events = set()
        self.unused_activities = set()
        # compute shortest s-t-path with upper_bounds
        start_event_id = ean.get_event_od(od_pair.get_origin(), od_pair.get_destination(), 'source', 1,
                                          1).get_event_id()
        end_event_id = ean.get_event_od(od_pair.get_origin(), od_pair.get_destination(), 'target', 1,
                                        None).get_event_id()
        self.length_sp_upper_bound = nx.dijkstra_path_length(self.ean_graph, start_event_id, end_event_id,
                                                        weight='upper_bound')

        # compute shortest s-i /i-t paths for all nodes
        source_to_intermediate = nx.single_source_dijkstra_path_length(self.ean_graph, start_event_id,
                                                                       weight='lower_bound')
        intermediate_to_target = nx.single_source_dijkstra_path_length(self.reverse_graph, end_event_id,
                                                                       weight='lower_bound')
        # Compute unused events
        for event in ean.get_events_network():
            event_id = event.get_event_id()
            if not nx.has_path(self.ean_graph, start_event_id, event_id) or not nx.has_path(self.reverse_graph,
                                                                                            end_event_id, event_id):
                self.unused_events.add(event)
                continue
            if source_to_intermediate[event_id] + intermediate_to_target[event_id] > self.length_sp_upper_bound:
                self.unused_events.add(event)

        # Compute unused activities
        for activity in self.activities_drive_wait_trans:
            if activity in PreprocessingTimPass.unused_trans_activities:
                self.unused_activities.add(activity)
                continue
            left_event_id = activity.get_left_event().get_event_id()
            right_event_id = activity.get_right_event().get_event_id()
            if not nx.has_path(self.ean_graph, start_event_id, left_event_id) or not nx.has_path(self.reverse_graph,
                                                                                                 end_event_id,
                                                                                                 right_event_id):
                self.unused_events.add(event)
                continue
            if source_to_intermediate[left_event_id] + self.ean_graph[left_event_id][right_event_id]['lower_bound'] \
                    + intermediate_to_target[right_event_id] > self.length_sp_upper_bound:
                self.unused_activities.add(activity)

        used_activities = set(ean.get_activities(['drive', 'wait', 'trans'])) - self.unused_activities
        used_events = set(ean.get_all_events()) - self.unused_events
        return [used_events, used_activities]
